import pathlib

import fastapi
from fastapi.responses import FileResponse, HTMLResponse, PlainTextResponse, RedirectResponse

from brigid.api import renderers
from brigid.api.sitemaps import build_sitemap_xml
from brigid.api.static_cache import cache
from brigid.api.utils import choose_language
from brigid.core import logging
from brigid.domain.urls import UrlsRoot
from brigid.library.storage import storage

router = fastapi.APIRouter()

logger = logging.get_module_logger()


####################
# Static files
####################


@router.get("/favicon.ico", response_model=None)
async def favicon() -> FileResponse | HTMLResponse:

    site = storage.get_site()

    if site.favicon is None:
        return HTMLResponse(content="")

    path = site.path.parent / site.favicon
    cache().set("/favicon.ico", path)

    return FileResponse(path, media_type="image/x-icon")


@router.get("/sitemap.xml")
async def site_map() -> PlainTextResponse:
    content = build_sitemap_xml()
    return PlainTextResponse(content, media_type="application/xml; charset=utf-8")


@router.get("/static/main.css")
async def page_css() -> FileResponse:
    css_file = pathlib.Path(__file__).parent.parent / "theme" / "static" / "main.css"

    cache().set("/static/main.css", css_file)

    return FileResponse(css_file, media_type="text/css")


@router.get("/static/posts/{article_slug}/{filename:path}")
async def static_file(request: fastapi.Request, article_slug: str, filename: str) -> FileResponse:
    article = storage.get_article(slug=article_slug)

    # TODO: could it be a security breach?
    path = article.path.parent / filename

    cache().set(request.url.path, path)

    # TODO: set media types according to file extension
    return FileResponse(path)


####################
# Technical routers
####################


@router.get("/{language}/feeds/atom")
async def feed_atom(language: str) -> HTMLResponse:
    return renderers.render_atom_feed(language)


# TODO: add RSS feed


@router.get("/robots.txt")
async def robots() -> PlainTextResponse:
    # language is not important here
    root_url = UrlsRoot(language="en")

    content = f"""\
User-agent: *
Sitemap: {root_url.to_site_map_full().url()}
"""

    return PlainTextResponse(content)


####################
# Content routers
####################


@router.get("/test-error")
async def test_error() -> HTMLResponse:
    1 / 0
    return HTMLResponse(content="This should not be shown")


@router.get("/")
async def root(request: fastapi.Request) -> RedirectResponse:
    language = choose_language(request)
    # TODO: show info to the user that language was chosen automatically
    return RedirectResponse(UrlsRoot(language=language).url(), status_code=302)


@router.get("/{language}")
async def blog_index(language: str) -> HTMLResponse:
    return renderers.render_index(language=language, raw_tags="")


@router.get("/{language}/tags")
async def tags_index_zero(language: str) -> RedirectResponse:
    return RedirectResponse(UrlsRoot(language=language).url(), status_code=301)


@router.get("/{language}/tags/{tags:path}")
async def tags_index(language: str, tags: str = "") -> HTMLResponse:
    return renderers.render_index(language=language, raw_tags=tags)


@router.get("/{language}/posts/{article_slug}")
async def page_article(language: str, article_slug: str) -> HTMLResponse:
    return renderers.render_page(language, article_slug)



import os
import pathlib
from typing import Any

from markupsafe import Markup

from brigid.domain import constants as domain_constants
from brigid.domain import request_context as d_request_context
from brigid.domain.urls import UrlsRoot
from brigid.library.entities import Page
from brigid.library.storage import Storage, storage
from brigid.renderer.markdown_render import render_page as markdown_render_page
from brigid.renderer.markdown_render import render_page_intro as markdown_render_page_intro
from brigid.renderer.markdown_render import render_text as markdown_render_text
from brigid.renderer.static_files import ImageInfo, files
from brigid.theme.default_translations import translations
from brigid.theme.settings import PhotoSwipe, settings
from brigid.theme.utils import jinjafilter, jinjaglobal


@jinjafilter
def upper_first(text: str) -> str:
    return text[0].upper() + text[1:]


@jinjafilter
def render_page(page: Page) -> str:
    return Markup(markdown_render_page(page=page).content)


@jinjafilter
def render_page_intro(page: Page) -> str:
    return Markup(markdown_render_page_intro(page=page).content)


@jinjafilter
def render_text(text: str) -> str:
    return Markup(markdown_render_text(text).content)


@jinjaglobal
def get_storage() -> Storage:
    return storage


@jinjaglobal
def photoswipe_settings() -> PhotoSwipe:
    return settings.photoswipe


@jinjaglobal
def image_info(path: pathlib.Path) -> ImageInfo:
    return files.image_info(path)


@jinjaglobal
def root_url(language: str) -> UrlsRoot:
    return UrlsRoot(language=language)


@jinjaglobal
def translate_tag(language: str, tag: str) -> str:
    site = storage.get_site()
    return site.languages[language].tags_translations[tag]


def _translate(language: str, text_id: str) -> str:
    site = storage.get_site()

    if text_id in site.languages[language].theme_translations:
        return site.languages[language].theme_translations[text_id]

    if language not in translations:
        language = site.default_language

    if language not in translations:
        language = "en"

    if text_id in translations[language]:
        return translations[language][text_id]

    return text_id


@jinjaglobal
def translate_theme(language: str, text_id: str) -> str:
    return Markup(_translate(language, text_id))


@jinjaglobal
def test_marker(marker: str) -> str:

    if not os.environ.get("BRIGID_TESTS_RUNNING"):
        return ""

    return Markup(marker)


@jinjaglobal
def brigid_repository() -> str:
    return domain_constants.brigid_repository


@jinjaglobal
def request_context_get(name: str) -> Any:
    return d_request_context.get(name)



from collections import defaultdict

from brigid.domain import request_context
from brigid.library.storage import storage
from brigid.renderer.markdown_render import render_page


class Connectivity:
    __slots__ = ("_connections", "_reverse_connections", "_processed_pages")

    def __init__(self) -> None:
        self._processed_pages: set[str] = set()
        self._connections: dict[str, set[str]] = defaultdict(set)
        self._reverse_connections: dict[str, set[str]] = defaultdict(set)

    def get_referenced_to(self, page_id: str) -> set[str]:
        return self._connections[page_id]

    def get_referenced_from(self, page_id: str) -> set[str]:
        return self._reverse_connections[page_id]

    def add_connection(self, target_page_id: str, reference_page_id: str):
        self._connections[reference_page_id].add(target_page_id)
        self._reverse_connections[target_page_id].add(reference_page_id)

    def process_page(self, page_id: str) -> None:
        if page_id in self._processed_pages:
            return

        with request_context.init():
            request_context.set("storage", storage)
            render_page(storage.get_page(page_id))

        self._processed_pages.add(page_id)

    def initialize(self) -> None:
        for language in storage.get_site().allowed_languages:
            for page in storage.get_posts(language=language):
                if page.id not in self._processed_pages:
                    self.process_page(page.id)


connectivity = Connectivity()



import datetime
import enum

from brigid.core.entities import BaseEntity


# TODO: template names should have prefixes
class Template(enum.StrEnum):
    article_page = "article.html.j2"
    index_page = "blog_index.html.j2"


class MetaInfo(BaseEntity):
    site_title: str

    language: str
    allowed_languages: list[str]

    title: str
    seo_description: str
    author: str
    tags: list[str]
    published_at: datetime.datetime | None

    seo_image_url: str | None



from typing import Any
from unittest import mock
from xml.dom import minidom  # noqa: S408

from brigid.domain import request_context
from brigid.domain.urls import UrlsTags
from brigid.theme.templates import env


def assert_compare_html(a: str, b: str):
    assert minidom.parseString(a).toprettyxml() == minidom.parseString(b).toprettyxml()  # noqa: S318


class TestTagsUrl:

    template = """
{% from "utils.html.j2" import tags_url %}
{{tags_url("xxx", current_url, "") }}
"""

    url = UrlsTags(language="en", page=13, required_tags=("a", "d"), excluded_tags=("c", "b", "e"))

    def render(self, context: dict[str, Any], template: str | None = None) -> str:
        if template is None:
            template = self.template

        return env().from_string(template).render(context)

    def test(self) -> None:
        context = {"current_url": self.url}

        request_context.set("language", "en")
        request_context.set("url", self.url)

        result = self.render(context)

        expected_result = """
  <a class=""
     href="http://0.0.0.0:8000/en/tags/a/-b/-c/d/-e/13">
    xxx
  </a>
 """

        assert_compare_html(result, expected_result)

    def test_render_class(self) -> None:
        template = """
{% from "utils.html.j2" import tags_url %}
{{tags_url("xxx", current_url, "aaa bbb") }}
"""

        context = {"current_url": self.url}

        request_context.set("language", "en")
        request_context.set("url", self.url)

        result = self.render(context, template=template)

        expected_result = """
  <a class="aaa bbb"
     href="http://0.0.0.0:8000/en/tags/a/-b/-c/d/-e/13">
    xxx
  </a>
 """

        assert_compare_html(result, expected_result)

    @mock.patch("brigid.domain.urls.UrlsTags.total_pages", 100500)
    def test_prev(self) -> None:
        context = {"current_url": self.url}

        page_url = self.url.move_page(1)

        assert self.url.is_prev_to(page_url)

        request_context.set("language", "en")
        request_context.set("url", page_url)

        result = self.render(context)

        expected_result = """
  <a class="" rel="prev" href="http://0.0.0.0:8000/en/tags/a/-b/-c/d/-e/13">
    xxx
  </a>
 """
        assert_compare_html(result, expected_result)

    @mock.patch("brigid.domain.urls.UrlsTags.total_pages", 100500)
    def test_next(self) -> None:
        context = {"current_url": self.url}

        page_url = self.url.move_page(-1)

        assert self.url.is_next_to(page_url)

        request_context.set("language", "en")
        request_context.set("url", page_url)

        result = self.render(context)

        expected_result = """
  <a class=""
     rel="next"
     href="http://0.0.0.0:8000/en/tags/a/-b/-c/d/-e/13">
    xxx
  </a>
 """

        assert_compare_html(result, expected_result)



import re

from brigid.library.entities import Page
from brigid.validation.entities import Error

UNORDERED_LIST_LINE_RE = re.compile(r"^\s*[-*+]\s+")
ORDERED_LIST_LINE_RE = re.compile(r"^\s*\d+\.\s+")


def is_list_line(line: str) -> bool:
    return bool(UNORDERED_LIST_LINE_RE.match(line)) or bool(ORDERED_LIST_LINE_RE.match(line))


def is_line_empty(line: str) -> bool:
    return not line.strip()


# TODO: what with nested lists?
# TODO: what with numbered lists?
# TODO: make difference for different list types (-, *, +, digit)
def parser(text: str) -> bool:  # noqa: CCR001

    is_in_list = False

    can_has_problem = False

    for line in text.split("\n"):
        if is_list_line(line):
            if can_has_problem:
                return True

            is_in_list = True
            continue

        if is_line_empty(line) and is_in_list:
            can_has_problem = True
            continue

        is_in_list = False
        can_has_problem = False

    return False


def page_has_correct_list_formatting(page: Page) -> list[Error]:
    if not parser(page.body):
        return []

    return [Error(filepath=page.path, message="Page has incorrect list formatting")]



import uuid

from brigid.library.storage import storage
from brigid.library.tests import make as library_make
from brigid.validation.global_validators import required_article


class TestPageIsRendered:

    def test_article_does_not_exists(self) -> None:
        validator = required_article(slug=uuid.uuid4().hex)
        assert validator() != []

    def test_has_article(self) -> None:
        article = library_make.article()

        for language in storage.get_site().allowed_languages:
            library_make.page(article=article, language=language)

        validator = required_article(slug=article.slug)
        assert validator() == []

    def test_has_no_page_in_one_language(self) -> None:

        site = storage.get_site()

        article = library_make.article()

        assert len(site.allowed_languages) > 1

        for language in list(storage.get_site().allowed_languages)[1:]:
            library_make.page(article=article, language=language)

        validator = required_article(slug=article.slug)
        assert validator() != []



import datetime
import pathlib
import uuid
from typing import Iterable

from brigid.core.utils import now
from brigid.library.entities import Article, ArticleType, Page
from brigid.library.storage import storage


def article(
    path: pathlib.Path | None = None, slug: str | None = None, type: ArticleType = ArticleType.post
) -> Article:

    if path is None:
        path = pathlib.Path("/tmp/brigid/tests") / uuid.uuid4().hex / "article.toml"  # noqa: S108

    if slug is None:
        slug = uuid.uuid4().hex

    article = Article(path=path, slug=slug, type=type, pages={}, tags=set())

    storage.add_article(article)

    return article


_article = article


def page(  # noqa: CFQ002
    article: Article | None = None,
    path: pathlib.Path | None = None,
    published_at: datetime.datetime | None = None,
    language: str = "en",
    title: str | None = None,
    seo_description: str | None = None,
    seo_image: str | None = None,
    body: str | None = None,
    tags: Iterable[str] = (),
    template: str | None = None,
) -> Page:

    if article is None:
        article = _article()

    if path is None:
        path = article.path.parent / f"{language}.md"

    if title is None:
        title = f"Title: {article.slug} {language}"

    if seo_description is None:
        seo_description = f"Description: {article.slug} {language}"

    if published_at is None:
        published_at = now()

    if body is None:
        body = f"Body: {article.slug} {language}"

    page = Page(
        article_id=article.id,
        path=path,
        published_at=published_at,
        language=language,
        title=title,
        seo_description=seo_description,
        seo_image=seo_image,
        body=body,
        tags=set(tags),
        template=template,
    )

    storage.add_page(page)

    return page



import pytest

from brigid.domain.html import strip_html


class TestStripHtml:

    @pytest.mark.parametrize(
        "text_input,text_expected",
        [
            ["<b>test</b>", "test"],
            ["", ""],
            ["tst test", "tst test"],
            ["complex <p>test</p> bla-bla <<<", "complex test bla-bla <<<"],
            ['complex with "quotes" test', 'complex with "quotes" test'],
        ],
    )
    def test_strip(self, text_input: str, text_expected: str) -> None:
        assert strip_html(text_input) == text_expected



import functools
import pathlib
import re

import frontmatter
import toml

from brigid.core import logging
from brigid.library.entities import Article, Collection, Page, Redirects, Site, SiteLanguage
from brigid.library.storage import storage

logger = logging.get_module_logger()


def log_error(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            logger.error(f"error_while_{func.__name__}", args=args, kwargs=kwargs)
            raise

    return wrapper


class FrontmatterTOMLHandler(frontmatter.TOMLHandler):
    FM_BOUNDARY = re.compile(r"^\-{3,}\s*$", re.MULTILINE)
    START_DELIMITER = END_DELIMITER = "---"


@log_error
def find_page_paths(directory: pathlib.Path) -> list[pathlib.Path]:  # noqa: CCR001

    site = storage.get_site()

    page_paths = []

    for root, _, files in directory.walk(follow_symlinks=True):
        for filename in files:
            if not filename.endswith(".md"):
                continue

            if filename[:-3] not in site.allowed_languages:
                continue

            page_paths.append((root / filename).resolve().absolute())

    logger.info("found_pages", count=len(page_paths))

    return page_paths


@log_error
def find_article_paths(directory: pathlib.Path) -> list[pathlib.Path]:
    article_paths = []

    for root, _, files in directory.walk(follow_symlinks=True):
        for filename in files:
            if filename == "article.toml":
                article_paths.append((root / filename).resolve().absolute())

    logger.info("found_articles", count=len(article_paths))

    return article_paths


@log_error
def load_page(path: pathlib.Path) -> Page:

    path = path.resolve().absolute()

    article_path = path.parent / "article.toml"

    if storage.has_article(path=article_path):
        article = storage.get_article(path=article_path)

    elif not article_path.exists():
        raise NotImplementedError(f"article configs has not found at {article_path}")

    else:
        raise NotImplementedError(f"article has not found for {article_path}")

    with path.open() as f:
        page_data, page_content = frontmatter.parse(f.read(), handler=FrontmatterTOMLHandler())

    if "tags" not in page_data:
        page_data["tags"] = article.tags

    if page_data["seo_image"] == "":
        page_data["seo_image"] = None

    page = Page(
        article_id=article.id,
        path=path,
        language=path.stem,
        body=page_content,
        **page_data,
    )

    return page


@log_error
def load_pages(directory: pathlib.Path) -> None:
    page_paths = find_page_paths(directory=directory)

    for page_path in page_paths:
        page = load_page(page_path)
        storage.add_page(page)


@log_error
def load_article(path: pathlib.Path) -> Article:
    path = path.resolve().absolute()

    article_content = path.read_text()

    article_data = toml.loads(article_content)
    article = Article(path=path, **article_data)

    return article


@log_error
def load_articles(directory: pathlib.Path) -> None:
    article_paths = find_article_paths(directory=directory)

    for article_path in article_paths:
        article = load_article(article_path)
        storage.add_article(article)


# TODO: exclude 'site' directory from pages search
#       or freeze all highl-level directories?
@log_error
def load_site(directory: pathlib.Path) -> None:  # noqa: CCR001
    languages = {}
    site = None
    redirects = Redirects()

    for file_path in (directory / "site").glob("*.toml"):
        data = toml.loads(file_path.read_text())

        if file_path.name == "meta.toml":
            site = Site(**data, path=file_path.resolve().absolute())
            continue

        if file_path.name == "redirects.toml":
            redirects = Redirects(**data)
            continue

        languages[file_path.stem] = SiteLanguage(**data)

        for menu in languages[file_path.stem].menu:
            menu.language = file_path.stem

    if site is None:
        raise NotImplementedError("meta.toml has not found")

    if (directory / "site" / "footer.html").exists():
        site.footer_html = (directory / "site" / "footer.html").read_text()

    if (directory / "site" / "header.html").exists():
        site.header_html = (directory / "site" / "header.html").read_text()

    site.languages.update(languages)

    storage.set_site(site)

    storage.set_redirects(redirects)


@log_error
def load_collections(directory: pathlib.Path) -> None:
    for file_path in (directory / "collections").glob("*.toml"):
        data = toml.loads(file_path.read_text())

        path = file_path.resolve().absolute()

        collection = Collection(**data, path=path)

        storage.add_collection(collection)


@log_error
def load(directory: pathlib.Path) -> None:
    from brigid.library.connectivity import connectivity

    # goes fiest, because other loaders depend on it
    load_site(directory=directory)

    load_articles(directory=directory)
    load_pages(directory=directory)
    load_collections(directory=directory)

    connectivity.initialize()



import pytest

from brigid.library.tests import make as library_make
from brigid.validation.lists_formatting import is_line_empty, is_list_line, page_has_correct_list_formatting


@pytest.mark.parametrize(
    "line,is_list",
    [
        ["- item 1", True],
        ["    - item 2", True],
        ["* item 3", True],
        ["+ item 4", True],
        ["1. item 5", True],
        ["    1. item 6", True],
        ["", False],
        ["**bold text**", False],
        ["text", False],
        ["    text", False],
        ["-sddssd", False],
        ["    -sddssd", False],
        ["*sddssd", False],
        ["+sddssd", False],
        ["1.sddssd", False],
        ["    1.sddssd", False],
    ],
)
def test_is_list_line(line: str, is_list: bool) -> None:
    assert is_list_line(line) == is_list


@pytest.mark.parametrize(
    "line,is_empty",
    [
        ["", True],
        [" ", True],
        ["    ", True],
        ["text", False],
        ["    text", False],
        ["    - item 1", False],
        ["    * item 2", False],
        ["    + item 3", False],
        ["    1. item 4", False],
    ],
)
def test_is_line_empty(line: str, is_empty: bool) -> None:
    assert is_line_empty(line) == is_empty


class TestPageHasCorrectListFormatting:

    def test_empty_body(self) -> None:
        page = library_make.page(body="")
        assert page_has_correct_list_formatting(page) == []

    def test_complext_list_are_ok(self) -> None:
        body = """
text in paragraphs

- item 1
- item 2
    - item 3
    - item 4
- item 5
    * item 6
    * item 7
        - item 8
        - item 9
- item 10

text in paragraphs

1. item 1
2. item 2
    1. item 3
    2. item 4

text in paragraphs
        """

        page = library_make.page(body=body)
        assert page_has_correct_list_formatting(page) == []

    @pytest.mark.parametrize("prefix", ["-", "*", "+"])
    def test_item_elemsts_are_separated_by_newlines__unordered(self, prefix: str) -> None:
        body = f"""
{prefix} item 1
{prefix} item 2

{prefix} item 3
{prefix} item 4
        """

        page = library_make.page(body=body)
        assert page_has_correct_list_formatting(page) != []

    def test_item_elemsts_are_separated_by_newlines__ordered(self) -> None:
        body = """
1. item 1
2. item 2

1. item 3
2. item 4
        """

        page = library_make.page(body=body)
        assert page_has_correct_list_formatting(page) != []

    def test_item_elemsts_are_separated_by_newline__nested(self) -> None:
        body = """
- item 1
    - item 2
    - item 3
- item 4
    - item 5
    - item 6

    - item 7
    - item 8
- item 9
        """

        page = library_make.page(body=body)
        assert page_has_correct_list_formatting(page) != []



import pathlib

import pydantic_settings

from brigid.core.settings import BaseSettings


class Settings(BaseSettings):
    directory: pathlib.Path = pathlib.Path.cwd() / "content"

    model_config = pydantic_settings.SettingsConfigDict(env_prefix="BRIGID_LIBRARY_")


settings = Settings()



import copy
import posixpath
from typing import Any, Iterable
from urllib.parse import urlparse, urlunparse

from brigid.domain import request_context


def _base_url() -> str:
    return request_context.get("storage").get_site().url  # type: ignore


def normalize_url(url: str) -> str:

    if "://" not in url:
        raise NotImplementedError("Not full url")

    schema, tail = url.split("://")

    while "//" in tail:
        tail = tail.replace("//", "/")

    url = f"{schema}://{tail}"

    parsed_url = urlparse(url)
    normalized_path = posixpath.normpath(parsed_url.path)

    if normalized_path == ".":
        normalized_path = ""

    result = urlunparse(parsed_url._replace(path=normalized_path))

    if result[-1] == "/":
        result = result[:-1]

    return result


class UrlsBase:
    __slots__ = ("language",)

    def is_noindex(self) -> bool:
        return False

    def __init__(self, language: str) -> None:
        self.language = language

    def url(self) -> str:
        raise NotImplementedError("url")

    def file_url(self, relative_path: str) -> str:
        raise NotImplementedError("file_url")

    def to_root(self) -> "UrlsRoot":
        return UrlsRoot(language=self.language)

    def to_feeds_atom(self) -> "UrlsFeedsAtom":
        return UrlsFeedsAtom(language=self.language)

    def to_author(self) -> "UrlsAuthor":
        return UrlsAuthor(language=self.language)

    def to_post(self, slug: str) -> "UrlsPost":
        return UrlsPost(language=self.language, slug=slug)

    def to_filter(self, page: int = 1, require: Iterable[str] = (), exclude: Iterable[str] = ()) -> "UrlsTags":
        return UrlsTags(
            language=self.language,
            page=page,
            required_tags=require,
            excluded_tags=exclude,
        )

    def to_language(self, language: str) -> Any:
        new_urls = copy.deepcopy(self)
        new_urls.language = language
        return new_urls

    def to_site_map_full(self) -> "UrlsSiteMapFull":
        return UrlsSiteMapFull(language=self.language)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False

        return self.language == other.language


class UrlsRoot(UrlsBase):
    __slots__ = ()

    def url(self) -> str:
        return normalize_url(f"{_base_url()}/{self.language}")


class UrlsAuthor(UrlsBase):
    __slots__ = ()

    def url(self) -> str:
        return normalize_url(f"{_base_url()}/{self.language}/about")


class UrlsFeedsAtom(UrlsBase):
    __slots__ = ()

    def url(self) -> str:
        return normalize_url(f"{_base_url()}/{self.language}/feeds/atom")


class UrlsSiteMapFull(UrlsBase):
    __slots__ = ()

    def url(self) -> str:
        return normalize_url(f"{_base_url()}/sitemap.xml")


class UrlsPost(UrlsBase):
    __slots__ = ("slug",)

    def __init__(self, slug: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.slug = slug

    def url(self) -> str:
        return normalize_url(f"{_base_url()}/{self.language}/posts/{self.slug}")

    def file_url(self, relative_path: str) -> str:
        return normalize_url(f"{_base_url()}/static/posts/{self.slug}/{relative_path}")

    def __eq__(self, other: Any) -> bool:
        if not super().__eq__(other):
            return False

        return self.slug == other.slug


class UrlsTags(UrlsBase):
    __slots__ = ("page", "required_tags", "excluded_tags", "selected_tags", "_total_pages")

    def __init__(
        self,
        page: int,
        required_tags: Iterable[str],
        excluded_tags: Iterable[str],
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.page = page
        self.required_tags = frozenset(required_tags)
        self.excluded_tags = frozenset(excluded_tags)
        self.selected_tags = self.required_tags | self.excluded_tags
        self._total_pages: None | int = None

    def __eq__(self, other: Any) -> bool:
        if not super().__eq__(other):
            return False

        return (
            self.page == other.page
            and self.required_tags == other.required_tags
            and self.excluded_tags == other.excluded_tags
        )

    def _is_same_index(self, current_url: UrlsBase) -> bool:
        if not isinstance(current_url, UrlsTags):
            return False

        if self.required_tags != current_url.required_tags:
            return False

        if self.excluded_tags != current_url.excluded_tags:
            return False

        if self.language != current_url.language:
            return False

        return True

    def is_prev_to(self, current_url: UrlsBase) -> bool:
        return self._is_same_index(current_url) and self.page + 1 == current_url.page  # type: ignore

    def is_next_to(self, current_url: UrlsBase) -> bool:
        return self._is_same_index(current_url) and self.page - 1 == current_url.page  # type: ignore

    def _get_total_pages(self) -> int:
        storage = request_context.get("storage")  # type: ignore
        posts_per_page = storage.get_site().posts_per_page  # type: ignore

        all_pages = storage.get_posts(
            language=self.language,
            require_tags=self.required_tags,
            exclude_tags=self.excluded_tags,
        )

        return (len(all_pages) + posts_per_page - 1) // posts_per_page

    @property
    def total_pages(self) -> int:
        if self._total_pages is not None:
            return self._total_pages

        self._total_pages = self._get_total_pages()

        return self._total_pages

    def is_noindex(self) -> bool:
        # we do not want crawlers to index pages with filters
        # because there are infinite number of them
        return bool(self.selected_tags)

    def url(self) -> str:
        tags = list(self.required_tags | self.excluded_tags)

        tags.sort()

        tags = [tag if tag in self.required_tags else f"-{tag}" for tag in tags]

        if self.page != 1:
            tags.append(str(self.page))

        if not tags:
            return normalize_url(f"{_base_url()}/{self.language}")

        tags_path = "/".join(tags)

        return normalize_url(f"{_base_url()}/{self.language}/tags/{tags_path}")

    def first_page(self) -> "UrlsTags":
        return UrlsTags(
            language=self.language,
            page=1,
            required_tags=self.required_tags,
            excluded_tags=self.excluded_tags,
        )

    def last_page(self) -> "UrlsTags":
        return UrlsTags(
            language=self.language,
            page=self.total_pages,
            required_tags=self.required_tags,
            excluded_tags=self.excluded_tags,
        )

    def move_page(self, delta: int) -> "UrlsTags":
        new_page = self.page + delta

        if new_page < 1 or new_page > self.total_pages:
            raise NotImplementedError("Wrong page number")

        return UrlsTags(
            language=self.language,
            page=new_page,
            required_tags=self.required_tags,
            excluded_tags=self.excluded_tags,
        )

    def require(self, *tags: str) -> "UrlsTags":
        return UrlsTags(
            language=self.language,
            page=1,
            required_tags=self.required_tags | set(tags),
            excluded_tags=self.excluded_tags - set(tags),
        )

    def exclude(self, *tags: str) -> "UrlsTags":
        return UrlsTags(
            language=self.language,
            page=1,
            required_tags=self.required_tags - set(tags),
            excluded_tags=self.excluded_tags | set(tags),
        )

    def remove(self, *tags: str) -> "UrlsTags":
        return UrlsTags(
            language=self.language,
            page=1,
            required_tags=self.required_tags - set(tags),
            excluded_tags=self.excluded_tags - set(tags),
        )



from markdown.inlinepatterns import IMAGE_LINK_RE, ImageInlineProcessor

from brigid.renderer.context import render_context


class ImageInlineValidator(ImageInlineProcessor):

    def handleMatch(self, m, data):
        context = render_context.get()

        context.add_error(failed_text=data, message="Classic markdown image syntax is not allowed")

        return (None, None, None)


__all__ = ["ImageInlineValidator", "IMAGE_LINK_RE"]



from brigid.library.storage import storage
from brigid.library.tests import make as library_make
from brigid.renderer.markdown_render import render_page


class TestExternalLink:

    def test_target_blank(self) -> None:

        text = """
[xxx](http://example.com) + [yyy](http://example.org)
        """

        article = library_make.article()
        page = library_make.page(article, body=text)

        result = render_page(page)

        assert (
            result.content
            == '<p><a href="http://example.com" target="_blank">xxx</a> + <a href="http://example.org" target="_blank">yyy</a></p>'  # noqa: E501
        )

    def test_no_schema(self) -> None:

        text = """
[xxx](example.com)
        """

        article = library_make.article()
        page = library_make.page(article, body=text)

        result = render_page(page)

        assert len(result.errors) == 1
        assert result.errors[0].message == "Specify schema/protocol for external links"

    def test_used_for_internal_links(self) -> None:

        site = storage.get_site()

        text = f"""
[xxx]({site.url})
        """

        article = library_make.article()
        page = library_make.page(article, body=text)

        result = render_page(page)

        assert len(result.errors) == 1
        assert result.errors[0].message == "Use internal links syntax to specify links to the same domain"



translations = {
    "ru": {
        "read_more": "Далее",
        "no_posts_found": "Ничего не найдено.",
        "pagination_first_page": "к новым",
        "pagination_last_page": "к старым",
        "tags_filter_header": "Нашли постов:",
        "article_similar_header": "Похожее",
        "last_posts_header": "Последние посты",
        "similarity_score_tooltip": "Уровень похожести",
        "license": "лицензия",
        "content_repository": "исходники постов",
        "site_is_created_with": "сайт создан с помощью",
        "icons_attribution": 'сайт использует <a href="https://github.com/tabler/tabler-icons" target="_blank">Tabler Icons</a>, лицензированные под <a href="https://github.com/tabler/tabler-icons/blob/main/LICENSE" target="_blank">лицензией MIT</a>',  # noqa
    },
    "en": {
        "read_more": "Read more",
        "no_posts_found": "Nothing found.",
        "pagination_first_page": "to new",
        "pagination_last_page": "to old",
        "tags_filter_header": "Found posts:",
        "article_similar_header": "Similar",
        "last_posts_header": "Last posts",
        "similarity_score_tooltip": "Similarity score",
        "license": "license",
        "content_repository": "content repository",
        "site_is_created_with": "site is created with",
        "icons_attribution": 'site is using <a href="https://github.com/tabler/tabler-icons" target="_blank">Tabler Icons</a>, licensed under the <a href="https://github.com/tabler/tabler-icons/blob/main/LICENSE" target="_blank">MIT License</a>',  # noqa
    },
}



from typing import Any

import fastapi
from fastapi.responses import RedirectResponse
from sentry_sdk import capture_exception

from brigid.api import renderers
from brigid.api.utils import choose_language
from brigid.domain import request_context as d_request_context
from brigid.library.storage import storage


async def process_404(request, _):
    redirects = storage.get_redirects()

    original_path = request.url.path

    # normalize path

    if not original_path.startswith("/"):
        original_path = "/" + original_path

    if original_path[-1] == "/":
        original_path = original_path[:-1]

    if original_path in redirects.permanent:
        return RedirectResponse(redirects.permanent[original_path], status_code=301)

    language = choose_language(request)

    with d_request_context.init():
        d_request_context.set("storage", storage)
        return renderers.render_page(language, "404", status_code=404)


async def remove_double_slashes(request: fastapi.Request, call_next: Any):
    path = request.url.path

    if "//" not in path:
        return await call_next(request)

    while "//" in path:
        path = path.replace("//", "/")

    return RedirectResponse(path, status_code=301)


async def remove_trailing_slash(request: fastapi.Request, call_next: Any):
    path = request.url.path

    if path != "" and path != "/" and path[-1] == "/":
        return RedirectResponse(path[:-1], status_code=301)

    return await call_next(request)


async def set_content_language(request: fastapi.Request, call_next: Any):
    response = await call_next(request)

    if response.status_code != 200:
        return response

    if "content-language" in response.headers:
        return response

    path = request.url.path

    for language in storage.get_site().allowed_languages:
        if path.startswith(f"/{language}/") or path == f"/{language}":
            response.headers["content-language"] = language
            return response

    return response


async def process_expected_error(request, error):
    capture_exception(error)

    language = choose_language(request)

    with d_request_context.init():
        d_request_context.set("storage", storage)
        return renderers.render_page(language, "500", status_code=500)


async def request_context(request: fastapi.Request, call_next: Any):
    with d_request_context.init():
        d_request_context.set("storage", storage)
        return await call_next(request)
